CountersActorBehavior subclass: #CountersBoundedBuffer
	instanceVariableNames: 'array getIndex putIndex maxSize '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Counters-Ex'!


!CountersBoundedBuffer methodsFor: 'initialize'!

initialize: anInteger

	maxSize := anInteger.
	array := Array new: maxSize + 1.
	getIndex := 1.
	putIndex := 1! !

!CountersBoundedBuffer methodsFor: 'script'!

get

	| item |
	item := array at: getIndex.
	getIndex := (getIndex \\ array size) + 1.
	(Delay forSeconds: 2) wait.
	^item!

put: item

	array at: putIndex put: item.
	putIndex := (putIndex \\ array size) + 1.
	(Delay forSeconds: 2) wait! !

!CountersBoundedBuffer methodsFor: 'guards'!

completedGet
	"Number of completed gets.
	This abstract method will be useful to maintain consistency when counting
	elements in the buffer regardless of how many new get methods are defined.
	Otherwise we would have to redefine ALL guards each time."

	^self completed: #get!

guardOFget
	"Only one get at once and the buffer should be not empty."

	^(self current: #get) = 0
		and: [((self completed: #put:) - (self completedGet)) > 0]!

guardOFput: item
	"Only one put at once and the buffer should not be full."

	^(self current: #put:) = 0
		and: [((self completed: #put:) - (self completedGet)) < maxSize]! !

!CountersBoundedBuffer methodsFor: 'printing'!

printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	getIndex <= putIndex
		ifTrue:
			[getIndex to: putIndex - 1 do:
				[:i |
				(array at: i) printOn: aStream.
				aStream nextPut: $ ]]
		ifFalse:
			[getIndex to: array size do:
				[:i |
				(array at: i) printOn: aStream.
				aStream nextPut: $ ].
			1 to: putIndex -1 do:
				[:i |
				(array at: i) printOn: aStream.
				aStream nextPut: $ ]].
	aStream nextPut: $)! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

CountersBoundedBuffer class
	instanceVariableNames: ''!


!CountersBoundedBuffer class methodsFor: 'instance creation'!

new: size
	^self new initialize: size! !

!CountersBoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"CountersBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 2"

	| buffer producer consumer |
	buffer := (CountersBoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: 0) actor.
	consumer := (Consumer new buffer: buffer delay: 2) actor.
	producer runPut: numberItems.
	consumer runGet: numberItems! !

CountersBoundedBuffer subclass: #CountersGetRearBoundedBuffer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Counters-Ex'!


!CountersGetRearBoundedBuffer methodsFor: 'script'!

getRear
	putIndex := ((putIndex - 2) \\ (array size)) + 1.
	^array at: putIndex! !

!CountersGetRearBoundedBuffer methodsFor: 'guards'!

completedGet
	"Number of completed gets includes the new getRear method."

	^(super completedGet) + (self completed: #getRear)!

guardOFgetRear
	"Method getRear excludes itself as well as method put:.
	Moreover the buffer should have at least one more element than current number
	of activated get.
	Note that this leads to maximal concurrency:
	methods get and getRear may be active concurrently if the buffer has more than one element."

	^(self current: #getRear) = 0
		and: [((self current: #put) = 0)
			and: [((self completed: #put:) - (self completedGet)) > (self current: #get)]]!

guardOFput: item
	"As for superclass, except that we also have to exclude concurrent activity of getRear.
	Note that the guard of put: is maintained valid,
	thanks to the redefinition of abstract method completedGet."

	^(super guardOFput: item)
		and: [self current: #getRear = 0]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

CountersGetRearBoundedBuffer class
	instanceVariableNames: ''!


!CountersGetRearBoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"CountersGetRearBoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 1"

	| buffer producer consumer |
	buffer := (CountersGetRearBoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: 1) actor.
	consumer := (Consumer new buffer: buffer delay: ratio) actor.
	producer runPut: numberItems.
	consumer runGetRear: numberItems! !

CountersBoundedBuffer subclass: #CountersGet2BoundedBuffer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Counters-Ex'!


!CountersGet2BoundedBuffer methodsFor: 'script'!

get2

	^Array with: self get with: self get! !

!CountersGet2BoundedBuffer methodsFor: 'guards'!

completedGet
	"Number of completed gets includes the new get2 method."

	^(super completedGet) + ((self completed: #get2) * 2)!

guardOFget
	"As for superclass, except that we also have to exclude concurrent activity of get2.
	Note that the guard of get is maintained valid,
	thanks to the redefinition of abstract method completedGet."

	^super guardOFget
		and: [(self current: #get2) = 0]!

guardOFget2
	"Only one get or get2 at once and the buffer should contain at least one element."

	^(self current: #get2) = 0
		and: [((self current: #get) = 0)
			and: [((self completed: #put:) - (self completedGet)) > 1]]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

CountersGet2BoundedBuffer class
	instanceVariableNames: ''!


!CountersGet2BoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"CountersGet2BoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 2"

	| buffer producer consumer |
	buffer := (CountersGet2BoundedBuffer new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: ratio) actor.
	consumer := (Consumer new buffer: buffer delay: 1) actor.
	producer runPut: numberItems.
	consumer runGet2: numberItems! !

ActorBehavior subclass: #BookUser
	instanceVariableNames: 'n '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Counters-Ex'!
BookUser comment:
'This class has been initially designed and implemented by the DESS class at Universite de Nantes, advised by Jean Bezivin and Olivier Roux.'!


!BookUser methodsFor: 'accessing'!

n: anInteger
	n := anInteger! !

!BookUser methodsFor: 'printing'!

printOn: aStream
	super printOn: aStream.
	n printOn: aStream! !

!BookUser methodsFor: 'private routine'!

reflechit: temps 
	Transcript show: self printString , ' commence a reflechir (duree = ' , temps printString , ')'; cr.
	temps timesRepeat: [Processor yield].
	Transcript show: self printString , ' finit de reflechir' ; cr! !

BookUser subclass: #Reader
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Counters-Ex'!
Reader comment:
'This class has been initially designed and implemented by the DESS class at Universite de Nantes, advised by Jean Bezivin and Olivier Roux.'!


!Reader methodsFor: 'script'!

reflechit: temps1 litSur: unLivre pendant: temps2 
	"Je reflechis pendant temps1 puis je lis le livre unLivre pendant le 
	temps temps2"

	self reflechit: temps1.
	unLivre lit: numero pendant: temps2!

reflechitAlea: temps1 litSur: unLivre pendant: temps2 
	"Je reflechis pendant temps1 puis je lis le livre unLivre pendant le 
	temps temps2"

	self reflechit: (temps1 * 1000) rounded.
	unLivre lit: numero pendant: (temps2 * 1000) rounded! !

!Reader methodsFor: 'private routine'!

lit: livre pendant: temps 
	| n |
	n := 0.
	Transcript show: 'lecteur' , numero printString , ' commence a lire (duree = ',temps printString,')
'.
	(1 to: temps )
		do: 
			[:i | 
			n := n + i.
			Processor yield].
	Transcript show: 'lecteur' , numero printString , ' finit de lire
'! !

BookUser subclass: #Writer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Counters-Ex'!
Writer comment:
'This class has been initially designed and implemented by the DESS class at Universite de Nantes, advised by Jean Bezivin and Olivier Roux.'!


!Writer methodsFor: 'script'!

reflechit: temps1 ecritSur: unLivre pendant: temps2 
	"Je reflechis pendant temps1 puis j'ecris sur le livre unLivre pendant 
	temps2"

	self reflechit: temps1.
	unLivre ecrit: numero pendant: temps2!

reflechitAlea: temps1 ecritSur: unLivre pendant: temps2 
	"Je reflechis pendant temps1 puis j'ecris sur le livre unLivre pendant 
	temps2"

	self reflechit: (temps1 * 1000) rounded.
	unLivre ecrit: numero pendant: (temps2 * 1000) rounded! !

!Writer methodsFor: 'private routine'!

ecrit: livre pendant: temps 
	| n |
	n := 0.
	Transcript show: 'ecrivain' , numero printString , ' commence a ecrire (duree = ',temps printString,')
'.
	(1 to: temps )
		do: 
			[:i | 
			n := n + i.
			Processor yield].
	Transcript show: 'ecrivain' , numero printString , ' finit d''ecrire
'! !

CountersActorBehavior subclass: #Book
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Counters-Ex'!
Book comment:
'This class has been initially designed and implemented by the DESS class at Universite de Nantes, advised by Jean Bezivin and Olivier Roux.'!


!Book methodsFor: 'script'!

ecrit: aNumero pendant: temps
	"L'ecrivain numero aNumero m'ecrit pendant le temps temps"
	| n |
	n := 0.
	Transcript show: 'Debut ecriture de l''ecrivain ' , aNumero printString , ' (duree = ' , temps printString,')'; cr.
	(1 to: temps)
		do: 
			[:i | 
			n := n + i.
			Processor yield].
	Transcript show: 'Fin ecriture de l''ecrivain ' , aNumero printString; cr!

lit: aNumero pendant: temps 
	"Le lecteur numero aNumero me lit pendant le temps temps"
	| toto |
	toto := 0.
	Transcript show: 'Debut lecture du lecteur ' , aNumero printString , ' (duree = ' , temps printString ,')'; cr.
	(1 to: temps)
		do: 
			[:i | 
			toto := toto + i.
			Processor yield].
	Transcript show: 'Fin lecture du lecteur ' , aNumero printString; cr! !

!Book methodsFor: 'guards'!

guardOFecrit: aNumero pendant: temps
	^((self current: #ecrit:pendant:) = 0)
		and: [(self current: #lit:pendant:) = 0
			and: [(self pending: #lit:pendant:) = 0]]!

guardOFlit: aNumero pendant: temps
	^(self current: #ecrit:pendant:) = 0! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Book class
	instanceVariableNames: ''!


!Book class methodsFor: 'example'!

example
	"Book example"
	| book reader1 reader2 reader3 writer1 writer2 |
	book := Book new actor.
	reader1 := (Reader new n: 1) actor.
	reader2 := (Reader new n: 2) actor.
	reader3 := (Reader new n: 3) actor.
	writer1 := (Writer new n: 1) actor.
	writer2 := (Writer new n: 2) actor.
	reader1 reflechit: 1 litSur: book pendant: 5.
	reader2 reflechit: 6 litSur: book pendant: 4.
	reader3 reflechit: 20 litSur: book pendant: 8.
	writer1 reflechit: 3 ecritSur: book pendant: 4.
	writer2 reflechit: 10 ecritSur: book pendant: 15! !
